@@ -2,41 +2,53 @@ from Whittler.classes.NestedObjectPointer import NestedObjectPointer
 from Whittler.config import Config
 import textwrap
 import random
+from collections import Counter
 
 
 actions = {
+    "general" : {
+        "exit" : "Gracefully exit the program",
+        "shell" : "Drop into an IPython shell to interact with the dataset."
+    },
     "navigation" : {
         "show [[limit]]" : "Show the current data context, up to [limit] entries (shows all entries by default). Mutes "+\
-                           "results or table entries with 0 relevant results.",
+                           "irrelevant results or table entries with 0 relevant results.",
         "showall [[limit]]" : "Show the current data context, up to [limit] entries (shows all entries by default). Includes "+\
-                              "results or table entries with 0 relevant results.",
+                              "irrelevant results or table entries with 0 relevant results.",
         "dig [attr]" : "Dig into a specific data grouping category, either by attribute name, or by attribute id",
         "up" : "Dig up a level into the broader data grouping category",
         "top" : "Dig up to the top level",
         "dump [[limit]]" : "Display every relevant result in every category, up to [limit] entries (shows all by default)",
-        "dumpall [[limit]]" : "Display every result, both relevant and irrelevant, in every category, up to [limit] entries "+\
-                              "(shows all by default)",
-        #"search [searchstr]" : "Finds the thing", # todo
         "exit" : "Gracefully exit the program"
     },
     "data model interaction" : {
         "irrelevant [[id]]" : "Mark all elements in the current context, or those referenced by [id], as irrelevant results",
         "relevant [[id]]" : "Mark all elements in the current context, or those referenced by [id], as relevant results",
-        "group [id] [[attr]]" : "Using data science, group all results in the database by similarity to the attribute referenced "+\
-                                "by [id]. Or, if [id] points to a specific result, group by similarity to a specific attribute of "+\
-                                "the result referenced by [id].",
+        "filter [str] [attr]" : "Mark all results containing [str] in a particular attribute [attr] as irrelevant "+\
+                                "(case-insensitive)",
+        "invfilter [str] [attr]" : "Mark all results not containing [str] in a particular attribute [attr] as irrelevant "+\
+                                   "(case-insensitive)",
+        "find [str] [attr]" : "Find and display all results that contain [str] in their [attr] attribute (case-sensitive)",
+        "invfind [str] [attr]" : "Find and display all results that do not contain [str] in their [attr] attribute (case-sensitive)",
+        "search [str] [attr]" : "Create a new result group with all results that contain [str] in their [attr] "+\
+                                "attribute (case-sensitive)",
+        "invsearch [str] [attr]" : "Create a new result group with all results that do not contain [str] in their [attr] "+\
+                                   "attribute (case-sensitive)",
+        #"group [attr]" : "",
+        "fuzzygroup [id] [attr]" : "Using data science, group all results in the database by similarity to the attribute "+\
+                                   "[attr] of the result referenced by [id].",
+        "ungroup [id]" : "Remove the result referenced by [id] from the result group in the current view.",
         "game [[id]]" : "Play a game where individual results are presented one-by-one, and the user is asked whether "+\
                         "the result is relevant or not and why. Using this information, other similar results are also "+\
                         "eliminated in bulk. If [id] is specified, then the results presented are limited to the result "+\
-                        "group represented by the specified [id].",
-        "filter [str] [[attr]]" : "Mark all results containing [str] in a particular attribute as irrelevant (case-insensitive)"
+                        "group represented by the specified [id]."
     },
     "output" : {
-        "quiet [[attr]]" : "Suppress an attribute from being displayed when printing out raw result data",
-        "unquiet [[attr]]" : "Undo the suppression from an earlier quiet command",
-        "solo [[attr]]" : "Print only a single attribute's value when printing out raw result data",
-        "SOLO [[attr]]" : "Print ONLY a single attribute's value when printing out raw result data, with no context IDs or "+\
-                          "attribute value names",
+        "quiet [attr]" : "Suppress an attribute from being displayed when printing out raw result data",
+        "unquiet [attr]" : "Undo the suppression from an earlier quiet command",
+        "solo [attr]" : "Print only a single attribute's value when printing out raw result data",
+        "SOLO [attr]" : "Print ONLY a single attribute's value when printing out raw result data, with no context IDs or "+\
+                        "attribute value names",
         "unsolo" : "Disable solo mode. Note that this retains any attributes suppressed using the \"quiet\" command.",
         "sort [colname]" : "Sorts the displayed results by the value in the specified column. Use quotes if the column name "+\
                            "has a space in it.",
@@ -48,10 +60,10 @@ actions = {
                              "column name has a space in it.",
         "history" : "Print all commands that have been run in this session so far",
         "width [numchars]" : "Modify the maxiumum terminal width, in characters, that all output will be formatted to",
-        "exportjson [fname] [[id]]" : "Export all relevant results in JSON form at into the file [fname]. Optionally, limit "+\
-                                      "the output to the result set as referenced by [id].",
-        "export [fname] [[id]]" : "Export all relevant results in Python Pickle (serialized binary) form at into the file [fname]. "+\
-                                  "Optionally, limit the output to the result set as referenced by [id]."
+        "exportjson [fname] [[id]]" : "Export all relevant results in JSON form into the gzip-compressed file [fname]. "+\
+                                      "Optionally, limit the output to the result set as referenced by [id].",
+        "export [fname] [[id]]" : "Export all relevant results in Python Pickle (serialized binary) form at into the gzip-compressed "+\
+                                  "file [fname]. Optionally, limit the output to the result set as referenced by [id]."
     }
 }
 
@@ -88,14 +100,15 @@ def get_redirect_file():
 
 class _NoInput:
     pass
-def wprint(s=_NoInput,end="\n"):
+def wprint(s=_NoInput,end="\n",quiet=False):
     if s is _NoInput:
         s = ""
     if not redirect_file is None:
         redirect_file.write(str(s)+end)
     if not global_redirect_file is None:
         global_redirect_file.write(str(s)+end)
-    print(s,end=end)
+    if not quiet:
+        print(s,end=end)
 
 def winput(msg):
     user_input = input(msg)
@@ -157,7 +170,7 @@ def get_ptr_from_id_arg(resultdb, args, id_arg_position=0, quiet=False):
             # The attribute values will be the .value property of the Vertex object given by the last
             # get_by_index operation called on this pointer.
             if ptr.path[-1].value == choice:
-                return ptr
+                return ptr.copy()
         if not quiet:
             wprint(f"Could not recognize \"{choice}\" as one of the attributes of this dataset.\n")
         return None
@@ -171,9 +184,9 @@ def get_ptr_from_id_arg(resultdb, args, id_arg_position=0, quiet=False):
 
 # False means no value was supplied for this arg position, None means failed to find corresponding context pointer
 def get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=0):
-    ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=attr_arg_position, quiet=True)
-    if ptr is False:
-        return False
+    # ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=attr_arg_position, quiet=True)
+    # if ptr is False:
+    #     return False
     attrname = args[attr_arg_position]
     if attrname not in resultdb.result_class.ATTRIBUTES:
         return None
@@ -190,24 +203,38 @@ def select_attribute(resultdb, msg):
         return None
     return attrs[index]
 
-def group_interactive(resultdb, groupattr, groupval):
+def group_interactive(resultdb, groupattr, groupval, max_print_count=10, max_print_chars_per_attrval=1000):
     wprint("Creating a result group set based on this attribute value...")
     SIMILARITY_THRESHOLD = resultdb.Config.SIMILARITY_THRESHOLD
+    wprint("Using similarity threshold {:.2f}...".format(SIMILARITY_THRESHOLD))
+    result_similarities = resultdb.find_similar_results(groupattr, groupval)
     while True:
-        print(SIMILARITY_THRESHOLD)
-        similar_results = resultdb.find_similar_results(groupattr, groupval, similarity_threshold=SIMILARITY_THRESHOLD)
-        wprint()
-        for val in sorted(set([res[groupattr] for res in similar_results])):
-            wprint(f"   {val.strip()}")
-        wprint()
+        similar_results = list(res for res,similarity in result_similarities if similarity>SIMILARITY_THRESHOLD)
+        attrvals_of_similar_results = Counter()
+        for res in similar_results:
+            attrval = dict.__getitem__(res,groupattr)
+            attrvals_of_similar_results[attrval] += 1
+        ordered_attrvals_of_similar_results = attrvals_of_similar_results.most_common()[:max_print_count]
+        for attrval_mcs,count in ordered_attrvals_of_similar_results:
+            attr_printstr = attrval_mcs.value
+            if len(attr_printstr) > max_print_chars_per_attrval:
+                attr_printstr = attr_printstr[:max_print_chars_per_attrval] + " [...]"
+            wprint(f"----- {count} result{'s' if count>1 else ''} with value:\n{attr_printstr}\n")
+        if max_print_count < len(attrvals_of_similar_results):
+            wprint(f"\n... and {sum(attrvals_of_similar_results.values())-sum(count for attrval,count in ordered_attrvals_of_similar_results)} more ...\n")
         wprint(f"I found {len(similar_results)} similar results with \"{groupattr}\" values as shown above.")
-        wprint("I want to create a result group set based on these findings, but I want to make sure it's OK.")
+        wprint("I want to create a result group set based on these findings. Is this grouping...")
+        wprint("   OK? (1)")
+        wprint("   too conservative? (2)")
+        wprint("   too liberal? (3)")
+        wprint("   never mind, abandon group creation (4)")
+        wprint("   not sure, show more results (5)")
         while True:
-            grouping_choice = winput("Is this grouping OK (1), too conservative (2), or too liberal (3)? Or, just abandon the group creation (4)? ")
+            grouping_choice = winput("> ")
             bad_input = False
             try:
                 grouping_choice = int(grouping_choice)
-                if grouping_choice not in [1,2,3,4]:
+                if grouping_choice not in [1,2,3,4,5]:
                     bad_input = True
             except:
                 bad_input = True
@@ -219,15 +246,22 @@ def group_interactive(resultdb, groupattr, groupval):
             pass
         elif grouping_choice == 2:
             SIMILARITY_THRESHOLD -= SIMILARITY_THRESHOLD / Config.SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
+            wprint("Using similarity threshold {:.2f}...".format(SIMILARITY_THRESHOLD))
             continue
         elif grouping_choice == 3:
             SIMILARITY_THRESHOLD += (1 - SIMILARITY_THRESHOLD) / Config.SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
+            wprint("Using similarity threshold {:.2f}...".format(SIMILARITY_THRESHOLD))
             continue
         elif grouping_choice == 4:
             wprint("OK, abandoning group creation.")
             break
+        elif grouping_choice == 5:
+            max_print_count *= 2
+            max_print_chars_per_attrval *= 2
+            continue
         resultdb.register_grouped_results(groupattr, groupval, similar_results)
         wprint("Created result group based on this entry. You can view this group in the \"groups\" pane in the top context.")
+        wprint("(You might have to use the \"showall\" command, since all results in this group are now marked irrelevant.)")
         ptr = NestedObjectPointer(resultdb)
         ptr.access_property("grouped_results")
         ptr.get_by_index(groupval)
@@ -353,8 +387,8 @@ def play_elimination_game(resultdb, obj):
         wprint("I can try to find other results similar to this one through various heuristics, and mark them irrelevant as well.\n")
 
         search_actions = {
-            1 : "contains this specific value",
-            2 : "contains a substring of this value (case-insensitive)",
+            1 : "contains the specific value above",
+            2 : "contains a particular substring (case-insensitive)",
             3 : "loosely resembles this value (fuzzy string matching)",
             4 : "never mind"
         }
@@ -414,14 +448,14 @@ def play_elimination_game(resultdb, obj):
                 if group_ptr is None:
                     continue
                 irrelevant_resultdict = group_ptr.give_pointed_object()
-                num_results_eliminated = len(irrelevant_resultdict.all_result_objects())
+                num_results_eliminated = len([r for r in irrelevant_resultdict.all_result_objects()])
                 irrelevant_resultdict.mark_irrelevant()
             elif answer == 4:
                 generate_report = False
             
             if generate_report:
                 excitement = num_results_eliminated//100
-                report = f"\n > Eliminating {num_results_eliminated} results{'.' if not excitement else ''}{'!'*excitement}\n"
+                report = f"\n > Eliminating {num_results_eliminated} results{'.' if not excitement else '!'*excitement}\n"
                 if excitement >= 3:
                     report = report.upper()
                 wprint(report)
